/*-
 * Copyright (c) 2007 Rui Paulo <rpaulo@fnop.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * $P4: //depot/projects/soc2007/rpaulo-macbook/dev/asmc/asmc.c#1 $
 *
 */

/*
 * Driver for Apple's System Management Console (SMC).
 * SMC can be found on the MacBook, MacBook Pro and Mac Mini.
 *
 * Inspired by the Linux applesmc driver.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/bus.h>
#include <sys/systm.h>
#include <sys/types.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>
#include <sys/sbuf.h>

#include <machine/bus.h>
#include <sys/rman.h>
#include <machine/resource.h>

#include <dev/asmc/asmcvar.h>


/*
 * Device interface.
 */

static void	asmc_identify(driver_t *, device_t);
static int	asmc_probe(device_t);
static int	asmc_attach(device_t);
static int	asmc_detach(device_t);

/*
 * SMC functions.
 */
static int	asmc_init(device_t);
static int	asmc_wait(device_t, uint8_t);
static int	asmc_key_write(device_t, const char *, uint8_t *, uint8_t);
static int	asmc_key_read(device_t, const char *, uint8_t *, uint8_t);
static int	asmc_fan_count(device_t);
static int	asmc_fan_getvalue(device_t, const char *, int);
static int	asmc_temp_getvalue(device_t, const char *);
static int	asmc_sms_read(device_t, const char *, int16_t *);
static int	asms_sms_calibrate(device_t);

/*
 * Model functions.
 */
static int	asmc_mb_sysctl_fanspeed(SYSCTL_HANDLER_ARGS);
static int	asmc_mb_sysctl_fansafespeed(SYSCTL_HANDLER_ARGS);
static int	asmc_mb_sysctl_fanminspeed(SYSCTL_HANDLER_ARGS);
static int	asmc_mb_sysctl_fanmaxspeed(SYSCTL_HANDLER_ARGS);
static int	asmc_mb_sysctl_fantargetspeed(SYSCTL_HANDLER_ARGS);
static int	asmc_temp_sysctl(SYSCTL_HANDLER_ARGS);
static int	asmc_mb_sysctl_sms_x(SYSCTL_HANDLER_ARGS);
static int	asmc_mb_sysctl_sms_y(SYSCTL_HANDLER_ARGS);
static int	asmc_mb_sysctl_sms_z(SYSCTL_HANDLER_ARGS);

static struct asmc_model *asmc_match(device_t);

#define ASMC_SMS_FUNCS	asmc_mb_sysctl_sms_x, asmc_mb_sysctl_sms_y, \
			asmc_mb_sysctl_sms_z

#define ASMC_FAN_FUNCS	asmc_mb_sysctl_fanspeed, asmc_mb_sysctl_fansafespeed, \
			asmc_mb_sysctl_fanminspeed, \
			asmc_mb_sysctl_fanmaxspeed, \
			asmc_mb_sysctl_fantargetspeed

struct asmc_model asmc_models[] = {
	{ 
	  "MacBook1,1", "Apple SMC MacBook Core Duo",
	  ASMC_SMS_FUNCS, ASMC_FAN_FUNCS,
	  ASMC_MB_TEMPS, ASMC_MB_TEMPNAMES, ASMC_MB_TEMPDESCS
	},

	{ 
	  "MacBook2,1", "Apple SMC MacBook Core 2 Duo",
	  ASMC_SMS_FUNCS, ASMC_FAN_FUNCS,
	  ASMC_MB_TEMPS, ASMC_MB_TEMPNAMES, ASMC_MB_TEMPDESCS
	},

	{ 
	  "MacBookPro1,1", "Apple SMC MacBook Pro Core Duo (15-inch)",
	  ASMC_SMS_FUNCS, ASMC_FAN_FUNCS,
	  ASMC_MB_TEMPS, ASMC_MB_TEMPNAMES, ASMC_MB_TEMPDESCS
	},

	{ 
	  "MacBookPro1,2", "Apple SMC MacBook Pro Core Duo (17-inch)",
	  ASMC_SMS_FUNCS, ASMC_FAN_FUNCS,
	  ASMC_MB_TEMPS, ASMC_MB_TEMPNAMES, ASMC_MB_TEMPDESCS
	},

	{ 
	  "MacBookPro2,1", "Apple SMC MacBook Pro Core 2 Duo (17-inch)",
	  ASMC_SMS_FUNCS, ASMC_FAN_FUNCS,
	  ASMC_MB_TEMPS, ASMC_MB_TEMPNAMES, ASMC_MB_TEMPDESCS
	},

	{ 
	  "MacBookPro2,2", "Apple SMC MacBook Pro Core 2 Duo (15-inch)",
	  ASMC_SMS_FUNCS, ASMC_FAN_FUNCS,
	  ASMC_MB_TEMPS, ASMC_MB_TEMPNAMES, ASMC_MB_TEMPDESCS
	},

	/* The Mac Mini has no SMS */
	{ 
	  "Macmini1,1", "Apple SMC Mac Mini",
	  NULL, NULL, NULL,
	  ASMC_FAN_FUNCS,
	  ASMC_MM_TEMPS, ASMC_MM_TEMPNAMES, ASMC_MM_TEMPDESCS
	},

	{ NULL, NULL }
};

#undef ASMC_SMS_FUNCS
#undef ASMC_FAN_FUNCS

/*
 * Driver methods.
 */
static device_method_t	asmc_methods[] = {
	DEVMETHOD(device_identify,	asmc_identify),
	DEVMETHOD(device_probe,		asmc_probe),
	DEVMETHOD(device_attach,	asmc_attach),
	DEVMETHOD(device_detach,	asmc_detach),

	{ 0, 0 }
};


static driver_t	asmc_driver = {
	"asmc",
	asmc_methods,
	sizeof(struct asmc_softc)
};

static devclass_t asmc_devclass;

DRIVER_MODULE(ascm, isa, asmc_driver, asmc_devclass, NULL, NULL);


static void
asmc_identify(driver_t *driver, device_t parent)
{
	if (device_find_child(parent, "asmc", -1) == NULL &&
	    asmc_match(parent))
		BUS_ADD_CHILD(parent, 0, "asmc", -1);
}


static struct asmc_model *
asmc_match(device_t dev)
{
	int i;
	char *model;

	model = getenv("smbios.system.product");

	for (i = 0; asmc_models[i].smc_model; i++) {
		if (!strncmp(model, asmc_models[i].smc_model, strlen(model))) {
			freeenv(model);
			return &asmc_models[i];
		}
	}

	freeenv(model);

	return NULL;
}

static int
asmc_probe(device_t dev)
{
	struct asmc_model *model;

	if (resource_disabled("asmc", 0))
		return ENXIO;

	model = asmc_match(dev);
	
	if (!model)
		return ENXIO;

	device_set_desc(dev, model->smc_desc);

	return BUS_PROBE_GENERIC;
}

static int
asmc_attach(device_t dev)
{
	int i, j;
	char name[2];
	struct asmc_softc *sc = device_get_softc(dev);
	struct asmc_model *model;

	model = asmc_match(dev);

	asmc_init(dev);

	sc->sc_model = model;

	/*
	 * hw.asmc tree.
	 */
	sysctl_ctx_init(&sc->sc_sysctl_ctx);
	sc->sc_root_tree = SYSCTL_ADD_NODE(&sc->sc_sysctl_ctx,
					   SYSCTL_STATIC_CHILDREN(_hw),
					   OID_AUTO,
					   device_get_name(dev),
					   CTLFLAG_RD, 0,
					   device_get_desc(dev));

	/*
	 * hw.asmc.fan.* tree.
	 */
	sc->sc_fan_tree[0] = SYSCTL_ADD_NODE(&sc->sc_sysctl_ctx,
					     SYSCTL_CHILDREN(sc->sc_root_tree),
					     OID_AUTO,
					     "fan",
					     CTLFLAG_RD, 0,
					     "Fan Root Tree");

	for (i = 1; i <= sc->sc_nfan; i++) {
		j = i - 1;
		name[0] = '0' + j;
		name[1] = 0;
		sc->sc_fan_tree[i] = 
			SYSCTL_ADD_NODE(&sc->sc_sysctl_ctx,
					SYSCTL_CHILDREN(sc->sc_fan_tree[0]),
					OID_AUTO,
					name,
					CTLFLAG_RD, 0,
					"Fan Subtree");

		SYSCTL_ADD_PROC(&sc->sc_sysctl_ctx,
				SYSCTL_CHILDREN(sc->sc_fan_tree[i]),
				OID_AUTO, "speed", CTLTYPE_INT | CTLFLAG_RD,
				dev, j, model->smc_fan_speed, "I",
				"Fan speed in RPM");

		SYSCTL_ADD_PROC(&sc->sc_sysctl_ctx,
				SYSCTL_CHILDREN(sc->sc_fan_tree[i]),
				OID_AUTO, "safespeed",
				CTLTYPE_INT | CTLFLAG_RD,
				dev, j, model->smc_fan_safespeed, "I",
				"Fan safe speed in RPM");

		SYSCTL_ADD_PROC(&sc->sc_sysctl_ctx,
				SYSCTL_CHILDREN(sc->sc_fan_tree[i]),
				OID_AUTO, "minspeed",
				CTLTYPE_INT | CTLFLAG_RD,
				dev, j, model->smc_fan_minspeed, "I",
				"Fan minimum speed in RPM");

		SYSCTL_ADD_PROC(&sc->sc_sysctl_ctx,
				SYSCTL_CHILDREN(sc->sc_fan_tree[i]),
				OID_AUTO, "maxspeed",
				CTLTYPE_INT | CTLFLAG_RD,
				dev, j, model->smc_fan_maxspeed, "I",
				"Fan maximum speed in RPM");

		SYSCTL_ADD_PROC(&sc->sc_sysctl_ctx,
				SYSCTL_CHILDREN(sc->sc_fan_tree[i]),
				OID_AUTO, "targetspeed",
				CTLTYPE_INT | CTLFLAG_RD,
				dev, j, model->smc_fan_targetspeed, "I",
				"Fan target speed in RPM");
	}

	/*
	 * hw.asmc.temp tree.
	 */

	sc->sc_temp_tree = SYSCTL_ADD_NODE(&sc->sc_sysctl_ctx,
					   SYSCTL_CHILDREN(sc->sc_root_tree),
					   OID_AUTO,
					   "temp",
					   CTLFLAG_RD, 0,
					   "Temperature sensors");

	for (i = 0; model->smc_temps[i]; i++) {
		SYSCTL_ADD_PROC(&sc->sc_sysctl_ctx,
				SYSCTL_CHILDREN(sc->sc_temp_tree),
				OID_AUTO, model->smc_tempnames[i],
				CTLTYPE_INT | CTLFLAG_RD,
				dev, i, asmc_temp_sysctl, "I",
				model->smc_tempdescs[i]);
	}

	if (model->smc_sms_x == NULL)
		goto out;

	/*
	 * hw.asmc.sms tree.
	 */
	sc->sc_sms_tree = SYSCTL_ADD_NODE(&sc->sc_sysctl_ctx,
					  SYSCTL_CHILDREN(sc->sc_root_tree),
					  OID_AUTO,
					  "sms",
					  CTLFLAG_RD, 0,
					  "Sudden Motion Sensor");
	SYSCTL_ADD_PROC(&sc->sc_sysctl_ctx,
			SYSCTL_CHILDREN(sc->sc_sms_tree),
			OID_AUTO, "x", CTLTYPE_INT | CTLFLAG_RD,
			dev, 0, model->smc_sms_x, "I",
			"Sudden Motion Sensor X value");

	SYSCTL_ADD_PROC(&sc->sc_sysctl_ctx,
			SYSCTL_CHILDREN(sc->sc_sms_tree),
			OID_AUTO, "y", CTLTYPE_INT | CTLFLAG_RD,
			dev, 0, model->smc_sms_y, "I",
			"Sudden Motion Sensor Y value");

	SYSCTL_ADD_PROC(&sc->sc_sysctl_ctx,
			SYSCTL_CHILDREN(sc->sc_sms_tree),
			OID_AUTO, "z", CTLTYPE_INT | CTLFLAG_RD,
			dev, 0, model->smc_sms_z, "I",
			"Sudden Motion Sensor Z value");

out:
	return 0;
}

static int
asmc_detach(device_t dev)
{
	struct asmc_softc *sc = device_get_softc(dev);

	sysctl_ctx_free(&sc->sc_sysctl_ctx);

	return 0;
}

static int
asmc_init(device_t dev)
{
	struct asmc_softc *sc = device_get_softc(dev);
	int i, error = 1;
	uint8_t buf[2];

	/*
	 * Wait up to 5 seconds for SMS initialization.
	 */
	for (i = 0; i < 100; i++) {
		if (asmc_key_read(dev, ASMC_KEY_SMS, buf, 2) == 0 && 
		    (buf[0] != 0x00 || buf[1] != 0x00)) {
			error = 0;
			goto out;
		}
	
		buf[0] = ASMC_SMS_INIT1;
		buf[1] = ASMC_SMS_INIT2;
		asmc_key_write(dev, ASMC_KEY_SMS, buf, 2);
		DELAY(50);
	}

out:
	asms_sms_calibrate(dev);

	sc->sc_nfan = asmc_fan_count(dev);
	if (sc->sc_nfan > ASMC_MAXFANS) {
		device_printf(dev, "more than %d fans were detected. Please "
			           "report this.", ASMC_MAXFANS);
		sc->sc_nfan = ASMC_MAXFANS;
	}			      

	return error;
}

/*
 * We need to make sure that the SMC acks the byte sent.
 * Just wait up to 100 ms.
 */
static int
asmc_wait(device_t dev, uint8_t val)
{
	u_int i;

	val = val & ASMC_STATUS_MASK;

	for (i = 0; i < 10000; i++) {
		if ((inb(ASMC_CMDPORT) & ASMC_STATUS_MASK) == val)
			return 0;
		DELAY(10);
	}

	device_printf(dev, "%s failed: 0x%x, 0x%x\n", __func__, val,
		      inb(ASMC_CMDPORT));

	
	return 1;

}

static int
asmc_key_read(device_t dev, const char *key, uint8_t *buf, uint8_t len)
{
	int i, error = 1;

	outb(ASMC_CMDPORT, ASMC_CMDREAD);
	if (asmc_wait(dev, 0x0c))
		goto out;

	for (i = 0; i < 4; i++) {
		outb(ASMC_DATAPORT, key[i]);
		if (asmc_wait(dev, 0x04))
			goto out;
	}

	outb(ASMC_DATAPORT, len);

	for (i = 0; i < len; i++) {
		if (asmc_wait(dev, 0x05))
			goto out;
		buf[i] = inb(ASMC_DATAPORT);
	}

	error = 0;
out:
	return error;
}

static int
asmc_key_write(device_t dev, const char *key, uint8_t *buf, uint8_t len)
{
	int i, error = -1;

	outb(ASMC_CMDPORT, ASMC_CMDWRITE);
	if (asmc_wait(dev, 0x0c))
		goto out;

	for (i = 0; i < 4; i++) {
		outb(ASMC_DATAPORT, key[i]);
		if (asmc_wait(dev, 0x04))
			goto out;
	}

	outb(ASMC_DATAPORT, len);

	for (i = 0; i < len; i++) {
		if (asmc_wait(dev, 0x04))
			goto out;
		outb(ASMC_DATAPORT, buf[i]);
	}

	error = 0;
out:
	return error;

}

/*
 * Fan control functions.
 */
static int
asmc_fan_count(device_t dev)
{
	uint8_t buf[1];

	if (asmc_key_read(dev, ASMC_KEY_FANCOUNT, buf, 1) < 0)
		return -1;

	return buf[0];
}

static int
asmc_fan_getvalue(device_t dev, const char *key, int fan)
{
	int speed;
	uint8_t buf[2];
	char fankey[5];

	snprintf(fankey, sizeof(fankey), key, fan);

	if (asmc_key_read(dev, fankey, buf, 2) < 0)
		return -1;

	speed = (buf[0] << 6) | (buf[1] >> 2);

	return speed;
}

static int
asmc_mb_sysctl_fanspeed(SYSCTL_HANDLER_ARGS)
{
	device_t dev = (device_t) arg1;
	int fan = arg2;
	int error;
	int32_t v;

	v = asmc_fan_getvalue(dev, ASMC_KEY_FANSPEED, fan);

	error = sysctl_handle_int(oidp, &v, 0, req);

	return error;
}

static int
asmc_mb_sysctl_fansafespeed(SYSCTL_HANDLER_ARGS)
{
	device_t dev = (device_t) arg1;
	int fan = arg2;
	int error;
	int32_t v;

	v = asmc_fan_getvalue(dev, ASMC_KEY_FANSAFESPEED, fan);

	error = sysctl_handle_int(oidp, &v, 0, req);

	return error;
}


static int
asmc_mb_sysctl_fanminspeed(SYSCTL_HANDLER_ARGS)
{
	device_t dev = (device_t) arg1;
	int fan = arg2;
	int error;
	int32_t v;

	v = asmc_fan_getvalue(dev, ASMC_KEY_FANMINSPEED, fan);

	error = sysctl_handle_int(oidp, &v, 0, req);

	return error;
}

static int
asmc_mb_sysctl_fanmaxspeed(SYSCTL_HANDLER_ARGS)
{
	device_t dev = (device_t) arg1;
	int fan = arg2;
	int error;
	int32_t v;

	v = asmc_fan_getvalue(dev, ASMC_KEY_FANMAXSPEED, fan);

	error = sysctl_handle_int(oidp, &v, 0, req);

	return error;
}

static int
asmc_mb_sysctl_fantargetspeed(SYSCTL_HANDLER_ARGS)
{
	device_t dev = (device_t) arg1;
	int fan = arg2;
	int error;
	int32_t v;

	v = asmc_fan_getvalue(dev, ASMC_KEY_FANTARGETSPEED, fan);

	error = sysctl_handle_int(oidp, &v, 0, req);

	return error;
}

/*
 * Temperature functions.
 */
static int
asmc_temp_getvalue(device_t dev, const char *key)
{
	int temp;
	uint8_t buf[2];

	if (asmc_key_read(dev, key, buf, 2) < 0)
		return -1;

	temp = buf[0] | ((buf[1] >> 6) * 25);

	return temp;
}

static int
asmc_temp_sysctl(SYSCTL_HANDLER_ARGS)
{
	device_t dev = (device_t) arg1;
	struct asmc_softc *sc = device_get_softc(dev);
	int error, val;

	val = asmc_temp_getvalue(dev, sc->sc_model->smc_temps[arg2]);

	error = sysctl_handle_int(oidp, &val, 0, req);

	return error;
}

/*
 * Sudden Motion Sensor functions.
 */
static int
asmc_sms_read(device_t dev, const char *key, int16_t *val)
{
	uint8_t buf[2];
	int error;

	switch (key[3]) {
	case 'X':
	case 'Y':
	case 'Z':
		error =	asmc_key_read(dev, key, buf, 2);
		break;
	default:
		device_printf(dev, "%s called with invalid argument %s\n",
			      __func__, key);
		error = 1;
		goto out;
	}

	*val = ((int16_t)buf[0] << 8) | buf[1];
out:
	return error;
}

static int
asms_sms_calibrate(device_t dev)
{
	struct asmc_softc *sc = device_get_softc(dev);

	asmc_sms_read(dev, ASMC_KEY_SMS_X, &sc->sms_rest_x);
	asmc_sms_read(dev, ASMC_KEY_SMS_Y, &sc->sms_rest_y);
	asmc_sms_read(dev, ASMC_KEY_SMS_Z, &sc->sms_rest_z);
}

static int
asmc_mb_sysctl_sms_x(SYSCTL_HANDLER_ARGS)
{
	device_t dev = (device_t) arg1;
	int error;
	int16_t val;
	int32_t v;

	asmc_sms_read(dev, ASMC_KEY_SMS_X, &val);

	v = (int32_t) val;
	error = sysctl_handle_int(oidp, &v, 0, req);

	return error;
}

static int
asmc_mb_sysctl_sms_y(SYSCTL_HANDLER_ARGS)
{
	device_t dev = (device_t) arg1;
	int error;
	int16_t val;
	int32_t v;

	asmc_sms_read(dev, ASMC_KEY_SMS_Y, &val);

	v = (int32_t) val;
	error = sysctl_handle_int(oidp, &v, 0, req);

	return error;
}

static int
asmc_mb_sysctl_sms_z(SYSCTL_HANDLER_ARGS)
{
	device_t dev = (device_t) arg1;
	int error;
	int16_t val;
	int32_t v;

	asmc_sms_read(dev, ASMC_KEY_SMS_Z, &val);

	v = (int32_t) val;
	error = sysctl_handle_int(oidp, &v, sizeof(v), req);

	return error;
}
